import json
import logging
from typing import Any

from django.conf import settings
from django.db.models import Model

DELIMITER = "."
LOGGER = logging.getLogger(__file__)

ENABLE_WEBSOCKET: bool = getattr(settings, "NANGO_WEBSOCKET_CONNECTION_DELAY", -1) > -1


def encode_pk(pk: Any) -> str:
    return json.dumps(pk)


def decode_pk(encoded_pk: str) -> Any:
    return json.loads(encoded_pk)


def instance_ref_to_channel_group_key(app_label: str, model: str, pk: Any) -> str:
    return ".".join([app_label, model, str(pk)])


def serialise_model_attr(instance: Model, attr: str) -> str:
    """
    Returns a string representation of the value of
    this instance.
    """
    value = getattr(instance, attr)
    return serialise_value(value=value)


def serialise_value(*, value: Any) -> str:
    """
    Returns a string representation of the value of
    this instance.
    """
    if isinstance(value, Model):
        return str(value.pk)
    return str(value)



import logging
from functools import partial
from inspect import isclass
from inspect import signature
from types import MethodType
from typing import Any
from typing import cast
from typing import Dict
from typing import Optional
from typing import Set

from django import forms
from django.conf import settings
from django.db.models import Model
from django.utils.html import format_html

from ..common import ENABLE_WEBSOCKET
from ..common import encode_pk
from ..common import serialise_model_attr

Fields = Set[str]

DEFAULT_DEBOUNCE_PERIOD: float = getattr(settings, "DEFAULT_DEBOUNCE_PERIOD", 0.2)

LOGGER = logging.getLogger(__file__)


class FormMixin:
    auto_clean_debounce_period: float = DEFAULT_DEBOUNCE_PERIOD
    auto_submit_debounce_period: float = DEFAULT_DEBOUNCE_PERIOD

    auto_clean_fields: Optional[Fields] = None
    auto_submit_fields: Optional[Fields] = None

    def __init__(self, *args: Any, **kw: Any) -> None:
        super().__init__(*args, **kw)
        if hasattr(self, "fields"):
            patch_widgets(form=cast("Form", self))

    def clean(self, *args: Any, **kw: Any) -> Any:
        set_original_form_values_on_instance(form=cast("Form", self))
        return super().clean(*args, **kw)  # type: ignore


class ModelForm(FormMixin, forms.ModelForm):  # type: ignore
    pass


class Form(FormMixin, forms.Form):
    pass


def set_original_form_values_on_instance(
    form: Form, *, instance: Optional[Model] = None
):
    """
    If the form has fields, look for the special __original-*
    fields. If provided, attach them to the associated instance,
    so they are respected when clean() is invoked
    """
    if not getattr(form, "fields", None):
        # form does not (yet) have fields, so nothing to do
        return

    for field_name, field in form.fields.items():
        field_key = (
            f"__original-{form.prefix}-{field_name}"
            if form.prefix
            else f"__original-{field_name}"
        )
        if field_key in form.data:
            original_value = form.data[field_key]
            try:
                (instance or form.instance)._original_form_values[  # type: ignore
                    field_name
                ] = original_value
            except AttributeError:
                model = (instance or form.instance).__class__  # type: ignore
                LOGGER.error(f"Model {model} is missing the nango mixin.")
                raise


def patch_widgets(form: Form) -> None:
    """
    Modify the widgets so they will display the hidden inputs alongside
    the normal stuff.
    """
    websocket_connection_delay = getattr(
        settings, "NANGO_WEBSOCKET_CONNECTION_DELAY", -1
    )
    assert hasattr(form, "fields")
    for field_name, field in form.fields.items():
        widget = field.widget

        def _render(
            self,
            template,
            context,
            *args,
            _original_render=widget._render,
            _field_name=field_name,
            **kw,
        ):
            result = _original_render(template, context, *args, **kw)
            if "widget" not in context:
                return result
            attrs: Dict[str, str] = {
                "data-app-label": form.instance._meta.app_label,
                "data-model-name": form.instance._meta.model_name,
                "data-instance-pk": encode_pk(form.instance.pk),
                "data-original-name": _field_name,
                "data-related-form-id": context["widget"]["attrs"]["id"],
                "type": "hidden",
                "name": "__original-" + context["widget"]["name"],
                "value": serialise_model_attr(form.instance, _field_name),
            }

            if ENABLE_WEBSOCKET:
                attrs["data-connection-delay"] = str(websocket_connection_delay)

                if _field_name in (form.auto_submit_fields or {}):
                    attrs["data-auto-submit-debounce-period"] = str(
                        getattr(
                            form, "auto_submit_debounce_period", DEFAULT_DEBOUNCE_PERIOD
                        )
                        * 1000
                    )

                if _field_name in (form.auto_clean_fields or {}):
                    form.auto_clean_fields
                    attrs["data-auto-clean-debounce-period"] = str(
                        getattr(
                            form, "auto_clean_debounce_period", DEFAULT_DEBOUNCE_PERIOD
                        )
                        * 1000
                    )

            hidden = format_html(
                "<input "
                + "\n\t".join(f"{attr}={str(value)!r}" for attr, value in attrs.items())
                + ">"
            )
            return result + hidden

        widget._render = MethodType(_render, widget)


def __getattr__(name: str) -> Any:
    original = getattr(forms, name)
    if isclass(original):
        # Anything derived from django's ModelForm gets
        # modified to have FormMixin too
        if issubclass(original, forms.Form) and not issubclass(original, FormMixin):

            return type(original.__name__, (FormMixin, original), {})
        return original

    if callable(original):
        sig = signature(original)
        # Anything with a 'form' argument which is derived from
        # django's ModelForm gets modified to have FormMixin too
        if form := sig.parameters.get("form"):
            default = form.default
            if issubclass(default, forms.Form) and not issubclass(default, FormMixin):
                return partial(
                    original, form=type(original.__name__, (FormMixin, default), {})
                )
        return original
    raise AttributeError(f"module {__name__} has no attribute {name}")



from django import forms
from django.contrib import admin
from nango import forms
from nango.common import ENABLE_WEBSOCKET


class AdminMixin:
    class Media:
        if ENABLE_WEBSOCKET:
            js = ["nango/ws.js", "nango/admin.js"]

    def save_model(self, request, obj, form, change):
        forms.set_original_form_values_on_instance(form=form, instance=obj)
        return super().save_model(request=request, obj=obj, form=form, change=change)

    def get_form(self, *args, **kw):
        form = super().get_form(*args, **kw)

        for attr in [
            "auto_clean_fields",
            "auto_submit_fields",
            "auto_clean_debounce_period",
            "auto_submit_debounce_period",
        ]:
            # Push this attr value from the generic
            # view to the form
            if attr_value := getattr(self, attr, None):
                setattr(form, attr, attr_value)

        return form


class ModelAdmin(AdminMixin, admin.ModelAdmin):
    form = forms.ModelForm



from inspect import isclass
from types import MethodType
from typing import Any

from django.views.generic import edit
from nango.db.models import TrackableMixin
from nango.forms import patch_widgets
from nango.forms import set_original_form_values_on_instance

"""
idea: in the template, if certain fields are selected to be live-tracked,
then have as_p append the hidden fields (later on) and invoke the js code
to register the items.

as_p() uses 'django/forms/p.html' (or 'django/forms/forsets/p.html') to render with,
so we can simply shadow it.

For now, leave this file here as other tricks might require python changes
"""

"""
from django.db.models import get_model
mymodel = get_model('some_app', 'SomeModel')
"""


class Mixin:
    def get_form(self):
        form = super().get_form()
        if not isinstance(form.instance, TrackableMixin):
            return form

        patch_widgets(form)
        original_clean = form.clean

        def new_clean(self, *args, **kw):
            set_original_form_values_on_instance(form=self)
            return original_clean(*args, **kw)

        form.clean = MethodType(new_clean, form)

        for attr in [
            "auto_clean_fields",
            "auto_submit_fields",
            "auto_clean_debounce_period",
            "auto_submit_debounce_period",
        ]:
            # Push this attr value from the generic
            # view to the form
            if attr_value := getattr(self, attr, None):
                setattr(form, attr, attr_value)

        return form


def __getattr__(name: str) -> Any:
    original_class = getattr(edit, name)
    if not isclass(original_class):
        raise AttributeError(f"module {__name__} has no attribute {name}")
    return type(
        original_class.__name__,
        (
            Mixin,
            original_class,
        ),
        {},
    )



from django.contrib.admin import site

from .admin import ModelAdmin


def register(model):
    """
    You can use this in place of admin.site.register
    with your nango-aware models
    """
    return site.register(model, ModelAdmin)



import json
import logging
from collections import defaultdict
from typing import Any
from typing import Dict
from typing import Set
from typing import Tuple

from asgiref.sync import sync_to_async
from channels.generic.websocket import AsyncWebsocketConsumer
from django.apps import apps
from django.core.exceptions import ValidationError

from .common import decode_pk
from .common import instance_ref_to_channel_group_key
from .common import serialise_value

LOGGER = logging.getLogger(__file__)

MessageType = Tuple[str, str, Any]


class LiveUpdatesConsumer(AsyncWebsocketConsumer):  # type: ignore
    def __init__(self, **kw):
        super().__init__(**kw)
        self._my_groups: Set[str] = set()
        self.fields_for_instance: Dict[MessageType, Dict[str, Any]] = defaultdict(dict)

    async def connect(self) -> None:
        await self.accept()

    def _retrieve_instance_values(self, app, model, pk, attrs):
        Model = apps.get_model(app, model)
        instance = Model.objects.get(pk=pk)
        return {attr: getattr(instance, attr) for attr in attrs}

    async def saved(self, info) -> None:
        message = info["message"]
        app = message["app"]
        model = message["model"]
        pkey = message["pk"]
        fields = self.fields_for_instance[(app, model, pkey)]
        instance = await sync_to_async(self._retrieve_instance_values)(
            app=app, model=model, pk=pkey, attrs=fields
        )
        for attr, original_value in fields.items():
            new_value = serialise_value(value=instance[attr])
            if original_value != new_value:
                message = dict(
                    message=dict(
                        action="modify",
                        app=app,
                        model=model,
                        pk=pkey,
                        attr=attr,
                        original_value=original_value,
                        new_value=new_value,
                    )
                )
                await self.send(text_data=json.dumps(message))
                fields[attr] = new_value

    async def receive(self, text_data: bytes) -> None:
        text_data_json = json.loads(text_data)
        action = text_data_json["action"]
        if action == "Register":
            await self.register(text_data_json=text_data_json)
            return
        if action == "Submit":
            await self.submit(data=text_data_json)
            return
        if action == "Clean":
            await self.clean(data=text_data_json)
            return
        assert False, f"unknown {action=}"

    async def clean(self, data: Any) -> None:
        await sync_to_async(self.clean_sync)(data=data)
        await self.send(text_data=json.dumps(dict(action="clean", message=data)))

    def clean_sync(self, data: Any) -> None:
        """
        Using the new value for this instance attribute,
        make a provisional change and return the cleaned
        version of the value to the ws client.

        If validation fails, instead return the
        validation messages.
        """
        dataset = data["dataset"]
        app_label = dataset["appLabel"]
        model_name = dataset["modelName"]
        pkey = decode_pk(dataset["instancePk"])
        Model = apps.get_model(app_label, model_name)
        try:
            try:
                instance = Model.objects.get(pk=pkey)
                setattr(instance, dataset["originalName"], data["currentValue"])
                instance.clean()
                data["cleanedValue"] = getattr(instance, dataset["originalName"])
            except ValidationError as exception:
                data["validationErrors"] = exception.messages
        except Exception as exception:
            data["error"] = str(exception)

    async def submit(self, data: Any) -> None:
        await self.send(text_data=json.dumps(dict(action="submit", message=data)))

    async def register(self, text_data_json: Any) -> None:
        for instance_ref in text_data_json["fields"]:
            app_label = instance_ref["appLabel"]
            model = instance_ref["model"]
            pkey = decode_pk(instance_ref["pk"])
            group_key = instance_ref_to_channel_group_key(
                app_label=app_label, model=model, pk=pkey
            )
            fields = self.fields_for_instance[(app_label, model, pkey)]
            fields[instance_ref["attr"]] = instance_ref["value"]
            self._my_groups.add(group_key)

        for group_key in self._my_groups:
            await self.channel_layer.group_add(
                channel=self.channel_name, group=group_key
            )
            await self.saved(
                info=dict(message=dict(app=app_label, model=model, pk=pkey))
            )

    async def disconnect(self, close_code: Any) -> None:
        for my_group in self._my_groups:
            await self.channel_layer.group_discard(
                group=my_group, channel=self.channel_name
            )



import logging

from django.urls import re_path

from . import consumers
from .common import ENABLE_WEBSOCKET

LOGGER = logging.getLogger(__file__)

websocket_urlpatterns = []

if ENABLE_WEBSOCKET:
    websocket_urlpatterns.append(
        re_path(r"ws/liveupdates/$", consumers.LiveUpdatesConsumer.as_asgi())
    )
else:
    LOGGER.debug("Nango websockets are disabled")



import logging
from inspect import isclass
from typing import Any
from typing import Dict
from typing import List

from asgiref.sync import async_to_sync

try:
    from channels.layers import get_channel_layer
except ModuleNotFoundError:
    get_channel_layer = None

from django.core.exceptions import ValidationError

from django.db import models
from django.db import transaction
from django.utils.translation import gettext_lazy as _

from ..common import DELIMITER
from ..common import serialise_model_attr

LOGGER = logging.getLogger(__file__)


class TrackableMixin:
    """
    Support value tracking. Required for CAS-style validation on
    submission of a form, as well as realtime WS client updates
    """

    def __init__(self, *args, **kw) -> None:
        super().__init__(*args, **kw)
        self._original_form_values: Dict[str, Any] = dict()

    class Meta:
        abstract = True

    def save(self, *args, **kw):
        result = super().save(*args, **kw)
        if get_channel_layer:
            channel_layer = get_channel_layer()
            model_identifier = DELIMITER.join(
                [self._meta.app_label, self._meta.model_name, str(self.pk)]
            )
            message = dict(
                type="saved",
                message=dict(
                    app=self._meta.app_label, model=self._meta.model_name, pk=self.pk
                ),
            )

            def on_commit() -> None:
                async_to_sync(channel_layer.group_send)(model_identifier, message)

            transaction.on_commit(on_commit)
        return result

    def clean(self, *args, **kw):
        # If there is no pkey yet, this is a new object, so can't conflict
        if self._original_form_values and self.pk:
            errors: List[ValidationError] = []
            copy = self.__class__.objects.select_for_update().get(pk=self.pk)
            for attr, expected_value in self._original_form_values.items():
                current_stored_value = serialise_model_attr(copy, attr)
                if current_stored_value != expected_value:
                    errors.append(
                        ValidationError(
                            _(
                                "Field %(attr)s was %(original)s when you started editing it, but is %(db)s in the database."
                            ),
                            params=dict(
                                attr=attr,
                                original=expected_value,
                                db=current_stored_value,
                            ),
                        )
                    )
            if errors:
                raise ValidationError(
                    [
                        ValidationError(
                            _(
                                "This object has changed while you were editing it. "
                                "You will need to reload this page and make your changes again, "
                                "if they are still appropriate."
                            )
                        )
                    ]
                    + errors
                )
        super().clean(*args, **kw)


class LockableMixin:
    """
    Support instance locking. Allows views to identify the instance
    as locked, preventing others from modifying it
    (buildout: optional per-field granularity)
    """

    class Meta:
        abstract = True


class Model(TrackableMixin, models.Model):
    class Meta:
        abstract = True


def __getattr__(name: str) -> Any:
    original = getattr(models, name)
    if isclass(original):
        if issubclass(original, models.Model) and not issubclass(
            original, TrackableMixin
        ):
            return type(original.__name__, (TrackableMixin, original), {})
    return original

    # raise AttributeError(f"module {__name__} has no attribute {name}")



from inspect import isclass
from typing import Any

from django import forms
from django.contrib import admin
from nango import forms

from . import site  # noqa
from .admin import *  # noqa


def __getattr__(name: str) -> Any:
    OriginalClass = getattr(admin, name)
    if not isclass(OriginalClass):
        raise AttributeError(f"module {__name__} has no attribute {name}")

    """
    if issubclass(OriginalClass, forms.ModelForm):

        class NewForm(forms.FormMixin, OriginalClass):  # type: ignore
            pass

        return NewForm
    """

    if Form := getattr(OriginalClass, "form", None):
        Form = OriginalClass.form
        if not issubclass(Form, forms.FormMixin):

            class NewForm(forms.FormMixin, Form):  # type: ignore
                pass

            Form = NewForm

        class NewAdmin(AdminMixin, OriginalClass):  # type: ignore
            form = Form

        return NewAdmin

    # no form attribute, so leave as-is
    return OriginalClass


def register(*args, **kw):
    """
    You can use this in place of admin.site.register
    with your nango-aware models
    """
    return admin.register(*args, **kw)



from os import environ

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def myversion():
    def clean_scheme(version):
        return environ["VERSION_STRING"]

    return {"local_scheme": clean_scheme}


setuptools.setup(
    use_scm_version=False,
    name="nango",
    version=environ.get("VERSION_STRING", "dev"),
    author="Nick Farrell",
    author_email="nicholas.farrell@gmail.com",
    description="Provide model integrity between requests",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nicois/nango",
    project_urls={
        "Bug Tracker": "https://github.com/nicois/nango/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "project"},
    packages=setuptools.find_packages(where="project"),
    python_requires=">=3.8",
    # package_data={"nango": ["*.html", "*.css", "*.js"]},
    include_package_data=True,
)
